from pathlib import Path
from copy import deepcopy
import click
import yaml
import os
from contextlib import contextmanager
from .repo import Repo, Remote
from .consts import gitcmd as git
from .tools import (
    _raise_error,
    safe_relative_to,
    is_empty_dir,
    _strip_paths,
    remember_cwd,
)
from .consts import REPO_TYPE_INT, REPO_TYPE_SUB


class Patchdir(object):
    def __init__(self, path, apply_from_here_dir):
        """
        Internal patches must be applied starting from subfolder.
        Patches from the main repo must be applied starting from parent main folder.
        """
        self._path = Path(path)
        self.apply_from_here_dir = apply_from_here_dir

    def __str__(self):
        return f"{self._path}"

    @property
    @contextmanager
    def path(self):
        with remember_cwd(self.apply_from_here_dir):
            path = self._path
            if self.apply_from_here_dir:
                path = self.apply_from_here_dir
            yield path


class Config(object):
    def __init__(
        self, force_type=None, recursive=False, common_vars=None, parent_config=None,
        force_gimera_file=None,
    ):
        self.force_type = force_type
        self._repos = []
        self.recursive = recursive
        self.parent_common_vars = common_vars
        self.parent_config = parent_config
        self.force_gimera_file = force_gimera_file
        self.load_config()

    @property
    def parent_path(self):
        if self.parent_config:
            return self.parent_config.config_file.parent
        else:
            return Path(".")

    @property
    def repos(self):
        return self._repos

    # @property
    # def root_path(self):
    #     p = self
    #     while p.parent_config:
    #         p = p.parent_config
    #     return p.config_file.parent

    def _get_config_file(self):
        if self.force_gimera_file:
            return self.force_gimera_file
        config_file = Path(os.getcwd()) / "gimera.yml"
        if not config_file.exists():
            _raise_error(f"Did not find: {config_file}")
        return config_file

    def load_config(self):
        self._repos = []
        self.config_file = self._get_config_file()

        self.yaml_config = yaml.load(
            self.config_file.read_text(), Loader=yaml.FullLoader
        )
        for repo in self.yaml_config.get("repos", []):
            repoitem = Config.RepoItem(self, repo)
            repoitem.collect_recursive_informations()

    def remove(self, path):
        config = yaml.load(self.config_file.read_text(), Loader=yaml.FullLoader)
        repos = config["repos"]
        repos2 = []
        for repo in repos:
            if (
                Path(repo["path"]).resolve().absolute()
                != Path(path).resolve().absolute()
            ):
                repos2.append(repo)
        config["repos"] = repos2
        self.config_file.write_text(yaml.dump(config, default_flow_style=False))

    def _store(self, repo, value):
        """
        Makes a commit of the changes.
        """
        main_repo = Repo(self.config_file.parent)
        if main_repo.staged_files:
            _raise_error("There mustnt be any staged files when updating gimera.yml")

        config = yaml.load(self.config_file.read_text(), Loader=yaml.FullLoader)
        param_repo = repo
        for repo in config["repos"]:
            if Path(repo["path"]) == param_repo.path:
                for k, v in value.items():
                    if k == "sha":
                        v = str(v)
                    repo[k] = v
                break
        else:
            config["repos"].append(value)
        for k, v in value.items():
            try:
                if getattr(param_repo, k) != v:
                    setattr(param_repo, k, v)
            except AttributeError as ex:
                raise Exception(f"Cannot set attribute {k}") from ex
        self.config_file.write_text(yaml.dump(config, default_flow_style=False))
        main_repo.please_no_staged_files()
        if self.config_file.resolve() in [
            x.resolve() for x in main_repo.all_dirty_files
        ]:
            main_repo.X(*(git + ["add", self.config_file]))
        if main_repo.staged_files:
            main_repo.X(*(git + ["commit", "-m", "auto update gimera.yml"]))

    def get_repos(self, names):
        if not names:
            return self.repos
        if isinstance(names, (Path, str)):
            names = [names]
        names = list(_strip_paths(names))
        res = []

        for item in self.repos:
            if not item.enabled:
                continue
            if str(item.path) in names:
                res.append(item)
                names.remove(str(item.path))
        if names:
            _raise_error(f"Invalid path: {','.join(names)}")
        return res

    class RepoItem(object):
        def __init__(self, config, config_section):
            """ """
            self.config = config
            self._sha = config_section.get("sha", None)
            self.enabled = config_section.get("enabled", True)
            self.freeze_sha = config_section.get("freeze_sha", False)
            self.path = Path(config_section["path"])
            self.branch = self.eval(str(config_section["branch"]))
            self.merges = config_section.get("merges", [])
            self.patches = config_section.get("patches", [])
            self.ignored_patchfiles = config_section.get("ignored_patchfiles", [])
            self.edit_patchfile = config_section.get("edit_patchfile", "")
            self._type = config_section["type"]
            self._url = self.eval(config_section["url"])
            self._remotes = config_section.get("remotes", {})
            if self.path in [x.path for x in config.repos]:
                _raise_error(f"Duplicate path: {self.path}")
            config._repos.append(self)
            self.internal_patch_dirs = []

            self.remotes = self._remotes.items() or None

            if self.merges:
                _merges = []
                for merge in self.merges:
                    remote, ref = merge.split(" ")
                    _merges.append((remote.strip(), ref.strip()))
                self.merges = _merges

            if self.type not in [REPO_TYPE_SUB, REPO_TYPE_INT]:
                _raise_error(
                    "Please provide type for repo "
                    f"{self.path}: either '{REPO_TYPE_INT}' or '{REPO_TYPE_SUB}'"
                )

        def ignore_patchfile(self, path):
            if path.name in [Path(x).name for x in self.ignored_patchfiles]:
                click.secho(
                    f"Warning: patchfile {path} is ignored and not applied. As configured.",
                    fg="yellow",
                )
                return True
            if self.edit_patchfile:
                if self.edit_patchfile_full_path == path:
                    return True

        def collect_recursive_informations(self):
            gimera_yml = self.config.config_file.parent / self.path / "gimera.yml"
            if not gimera_yml.exists():
                return
            config = yaml.load(gimera_yml.read_text(), Loader=yaml.FullLoader)
            internal_patch_dirs = config.get("common", {}).get("patches", [])
            self.internal_patch_dirs += internal_patch_dirs

        @property
        def common_vars(self):
            res = deepcopy(self.config.parent_common_vars or {})
            common = self.config.yaml_config.get("common", {})
            res.update(common.get("vars", {}))
            return res

        def eval(self, text):
            for k, v in self.common_vars.items():
                text = text.replace(f"${{{k}}}", str(v))
            if "${" in text:
                _raise_error(f"Please define variables for {text}")
            return text

        def drop_dead(self):
            Config().remove(self.path)

        @property
        def sha(self):
            return self._sha

        @sha.setter
        def sha(self, value):
            self._sha = value
            if not self.freeze_sha and os.getenv("GIMERA_NO_SHA_UPDATE") != "1":
                self.config._store(self, {"sha": value})

        def as_dict(self):
            return {
                "path": self.path,
                "branch": self.branch,
                "patches": self.patches,
                "type": self._type,
                "url": self._url,
                "merges": self.merges,
                "remotes": self._remotes,
            }

        @property
        def url(self):
            return self._url

        @url.setter
        def url(self, value):
            self._url = value

        @property
        def url_public(self):
            url = self._url.replace("ssh://git@", "https://")
            return url

        @property
        def type(self):
            if self.config.force_type:
                return self.config.force_type
            return self._type

        @type.setter
        def type(self, value):
            self._type = value

        def all_patch_dirs(self, rel_or_abs=None):
            if not rel_or_abs:
                raise ValueError("Please define rel_or_abs")

            def transform_outbound_patchdirs(dir):
                patch_path = Path(dir)
                apply_from_here = self.config.parent_path / self.path
                if rel_or_abs == "absolute":
                    root = self.config.config_file.parent
                    patch_path = root / patch_path
                    apply_from_here = root / apply_from_here
                dir = Patchdir(patch_path, apply_from_here)
                return dir

            if isinstance(self.patches, str):
                raise ValueError(
                    f"Patches must be a list. But is {self.patches} for {self.url}"
                )
            res = list(map(transform_outbound_patchdirs, self.patches))

            def transform_internal_patchdir(dir):
                patch_path = Path(self.eval(str(dir)))
                apply_from_here = self.path

                if rel_or_abs == "absolute":
                    root = self.config.config_file.parent
                    patch_path = root / self.path / patch_path
                    apply_from_here = root / apply_from_here
                dir = Patchdir(patch_path, apply_from_here)
                return dir

            res += list(map(transform_internal_patchdir, self.internal_patch_dirs))
            for test in res:
                with test.path as testpath:
                    if not (self.config.config_file.parent / testpath).exists():
                        click.secho(f"Warning: not found: {test}", fg="yellow")
            return res

        @property
        def fullpath(self):
            return self.config.config_file.parent / self.path

        def _get_type_of_patchfolder(self, path):
            for test in self.patches:
                if str(path).startswith(str(test)):
                    return "from_outside"

            for test in self.internal_patch_dirs:
                if str(path).startswith(self.eval(str(test))):
                    return "internal"
            raise ValueError(f"Undefined patchfolder: {path}")

        @property
        def edit_patchfile_full_path(self):
            ttype = self._get_type_of_patchfolder(self.edit_patchfile)
            if (ttype) == "from_outside":
                return self.config.config_file.parent / self.edit_patchfile
            elif ttype == "internal":
                return self.fullpath / self.edit_patchfile
            else:
                raise NotImplementedError(ttype)

        def abs(self, path):
            if str(path).startswith("/"):
                raise ValueError(f"Path must be relative: {path}")
            return self.config.config_file.parent / self.path / path



import subprocess
from .tools import get_url_type, reformat_url
import traceback
import os
import threading
import click
import uuid
from .submodule import _has_repo_latest_commit
from .consts import gitcmd as git
from .repo import Repo
from .cachedir import _get_cache_dir
from .tools import verbose
from .tools import wait_git_lock
from .tools import _raise_error
from .tools import try_rm_tree

threadLimiter = threading.BoundedSemaphore(4)


def _fetch_repos_in_parallel(
    main_repo, repos, update=None, minimal_fetch=None, no_fetch=None
):
    results = {"errors": {}, "urls": set()}
    if os.getenv("GIMERA_NON_THREADED", "0") == "1":
        threaded = False
    else:
        threaded = len(repos) > 1

    def _pull_repo(index, main_repo, repo_yml):
        threadLimiter.acquire()
        try:
            if repo_yml.url in results["urls"]:
                return
            verbose(f"Fetching {repo_yml.url}")
            results["urls"].add(repo_yml.url)
            cache_dir = _get_cache_dir(main_repo, repo_yml)
            repo = Repo(cache_dir)
            do_fetch = True
            if minimal_fetch:
                with wait_git_lock(cache_dir):
                    if repo_yml.sha:
                        if repo.contains(repo_yml.sha):
                            do_fetch = False
                    else:
                        if repo.contains_branch(repo_yml.branch):
                            do_fetch = False

            if do_fetch:
                with wait_git_lock(cache_dir):
                    _fetch_branch(
                        repo, repo_yml, filter_remote="origin", no_fetch=False
                    )

        except Exception as ex:
            if os.getenv("GIMERA_IGNORE_FETCH_ERRORS") == "1":
                click.secho(
                    "Following error is ignored because GIMERA_IGNORE_FETCH_ERRORS is set:"
                )
                click.secho(str(ex), fg="red")
            else:
                trace = traceback.format_exc()
                results["errors"][main_repo.path] = f"{ex}\n\n{trace}"
                if not threaded:
                    raise
        finally:
            threadLimiter.release()

    if not threaded:
        for index, repo in enumerate(repos):
            _pull_repo(index, main_repo, repo)
    else:

        threads = []
        for index, repo in enumerate(repos):
            t = threading.Thread(target=_pull_repo, args=(index, main_repo, repo))
            t.daemon = True
            threads.append(t)
        [x.start() for x in threads]
        [x.join() for x in threads]

        if results["errors"]:
            raise Exception(results["errors"])


def _fetch_branch(repo, repo_yml, no_fetch=False, filter_remote=None, **options):
    url = repo_yml.url

    fetch_exception = None
    if no_fetch:
        return
    for remote in repo.remotes:
        try:
            url = remote.url
            _set_url_and_fetch(
                repo, repo_yml, remote.name, url, filter_remote=filter_remote
            )
        except Exception as ex:
            fetch_exception = ex
            for combination in [
                ("git", "http"),
                ("http", "git"),
            ]:
                if get_url_type(url) == combination[0]:
                    url_http = reformat_url(url, combination[1])
                    try:
                        _set_url_and_fetch(
                            repo,
                            repo_yml,
                            remote.name,
                            url_http,
                            filter_remote=filter_remote,
                        )
                        break
                    except Exception:
                        raise fetch_exception
            else:
                raise fetch_exception

def _set_url_and_fetch(
    repo, repo_yml, remote_name, url, filter_remote=None, trycount=0
):
    repo.set_remote_url(remote_name, url)
    branch = repo_yml.branch
    todo_branches = [f"{branch}:{branch}"]
    success = False

    with wait_git_lock(repo.path):
        try:
            repo.out(*(git + ["fetch", remote_name] + todo_branches))
            success = True
        except subprocess.CalledProcessError:
            pass

    if success:
        if not _has_repo_latest_commit(repo, repo_yml.branch):
            success = False

    if not success:
        if trycount == 0:
            try_rm_tree(repo.path)
            _get_cache_dir(repo, repo_yml)
            _set_url_and_fetch(
                repo,
                repo_yml,
                remote_name,
                url,
                filter_remote=filter_remote,
                trycount=trycount + 1,
            )
        else:
            _raise_error(
                f"Even after rebuilding cache dir it was not possible to clone {repo_yml.path}"
            )




import inquirer
gitcmd = ["git", "-c", "protocol.file.allow=always"]
try:
    inquirer_theme = inquirer.themes.GreenPassion()
except:
    inquirer_theme = None

REPO_TYPE_INT = "integrated"
REPO_TYPE_SUB = "submodule"


import os
import click
import json
import shutil
import sys
from pathlib import Path
from .consts import gitcmd as git
from .repo import Repo
from .tools import prepare_dir
from .tools import remember_cwd

def _get_cache_dir(main_repo, repo_yml):
    url = repo_yml.url
    if not url:
        click.secho(f"Missing url: {json.dumps(repo_yml, indent=4)}")
        sys.exit(-1)
    path = Path(os.path.expanduser("~/.cache/gimera")) / url.replace(":", "_").replace(
        "/", "_"
    )
    path.parent.mkdir(exist_ok=True, parents=True)

    must_exist = ["HEAD", "refs", "objects", "config", "info"]
    if path.exists() and any(not (path / x).exists() for x in must_exist):
        shutil.rmtree(path)

    if not path.exists():
        click.secho(
            f"Caching the repository {repo_yml.url} for quicker reuse",
            fg="yellow",
        )
        with prepare_dir(path) as _path:
            with remember_cwd(
                "/tmp"
            ):  # called from other situations where path may not exist anymore
                Repo(main_repo.path).X(*(git + ["clone", "--bare", url, _path]))
    if repo_yml.sha:
        if not Repo(path).contain_commit(repo_yml.sha):
            # make a fetch quickly; sha is missing
            Repo(path).X(*(git + ["fetch", "--all"]))
    return path
